/*
 * Copyright (c) 2019 Exploratory Engineering
 */
#include <zephyr.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <net/socket.h>
#include <at_cmd.h>
#include <at_cmd_parser/at_cmd_parser.h>
#include <string.h>
#include <utils.h>

K_SEM_DEFINE(registered_on_network_sem, 0, 1);

#define PSM_ENBALE true
#define PSM_DISABLE false

/* Network register time */
#define REGISTER_ON_NETWORK_TIMEOUT_S (3*60)

/* PSM setting time after connection, specifically needed for NBIOT, which requires at least 30 seconds*/
#define PSM_MODE_SETTING_TIME_AFTER_CONNECTION_S (30*1000)

static void at_cmd_notification_handler(char *response) {
	utils_printk("Received AT Command Response%s\n", response);
	if (memcmp(response, "+CEREG", 6) != 0) {
		utils_printk("Response: %s\n", response);
		return;
	}

	struct at_param_list params = {0};
	if (at_params_list_init(&params, 1) != 0) {
		utils_printk("Error initializing AT param list.\n");
		return;
	}
	if (at_parser_params_from_str(&response[sizeof("+CEREG:")], NULL, &params) != 0) {
		utils_printk("Error parsing response: %s\n", response);
		utils_printk("params type: %d\n", params.params->type);
		utils_printk("params size: %d\n", params.params->size);
		utils_printk("params value: %d\n", params.params->value.int_val);
		utils_printk("params value: %s\n", params.params->value.str_val);
		goto cleanup;
	}
	char value[20];
	size_t value_length = 20;
	/*TODO: added workadound, because at_parser_params_from_str does not return short or int values, when using NSC v1.1.0*/
	if (at_params_string_get(&params, 0, value, &value_length) != 0) {
		utils_printk("Error getting short param: %s\n", response);
		utils_printk("params str value: %s\n", value);
		utils_printk("params str len: %d\n", value_length);
		goto cleanup;
	}
	if (memcmp(value, "1", 1) == 0) {
		k_sem_give(&registered_on_network_sem);
	}

cleanup:
	at_params_list_free(&params);
}


static bool subscribe_network_status_notifications() {
	if (at_cmd_write("AT+CEREG=1", NULL, 0, NULL) != 0) {
		utils_printk("Error subscribing to network status notifications.\n");
		return false;
	}
	at_cmd_set_notification_handler(at_cmd_notification_handler);
	return true;
}

bool systemmode_lte() {
	if (at_cmd_write("AT+CFUN=4", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT%XSYSTEMMODE=0,1,0,0", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT+CFUN=1", NULL, 0, NULL) != 0) {
		utils_printk("Error enabling system mode LTE: %d\n", errno);
		return false;
	}

	if (k_sem_take(&registered_on_network_sem, K_SECONDS(REGISTER_ON_NETWORK_TIMEOUT_S)) != 0) {
		utils_printk("systemmode_lte: Did not register on network within %d seconds.\n", REGISTER_ON_NETWORK_TIMEOUT_S);
		return false;
	}

	return true;
}

bool print_imei_imsi() {
	char resp[32] = {0};
	int err = at_cmd_write("AT+CGSN", resp, sizeof(resp), NULL);
	if (err != 0) {
		utils_printk("Error getting IMEI: %d.\n", err);
		return false;
	}
	resp[15] = 0;
	utils_printk("IMEI: %s\n", resp);

	err = at_cmd_write("AT+CIMI", resp, sizeof(resp), NULL);
	if (err != 0) {
		utils_printk("Error getting IMSI: %d.\n", err);
		return false;
	}
	resp[15] = 0;
	utils_printk("IMSI: %s\n", resp);

	return true;
}

static bool set_apn() {
	if (at_cmd_write("AT+CFUN=1", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT+CGATT=0", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT+CGDCONT=0,\"IP\",\"mda.ee\"", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT+CGDCONT?", NULL, 0, NULL) != 0 ||
		at_cmd_write("AT+CGATT=1", NULL, 0, NULL) != 0) {
		return false;
	}

	if (k_sem_take(&registered_on_network_sem, K_SECONDS(REGISTER_ON_NETWORK_TIMEOUT_S)) != 0) {
		utils_printk("set_apn: Did not register on network within %d seconds.\n", REGISTER_ON_NETWORK_TIMEOUT_S);
		return false;
	}

	return true;
}

static bool set_psm(bool psm_enable) {
	/*
		Requested_Periodic-TAU: 00101100  --> 12 hours
		Requested_Active-Time:  00000000  --> 0 seconds
	*/
	if(psm_enable == PSM_ENBALE)
	{
		utils_printk("Enable PSM\n");
		if (at_cmd_write("AT+CPSMS=1,,,\"00101100\",\"00000000\"", NULL, 0, NULL) != 0) {
			return false;
		}
		utils_printk("Enable PSM: Done\n");
	}
	else
	{
		utils_printk("Disable PSM\n");
		if (at_cmd_write("AT+CPSMS=0", NULL, 0, NULL) != 0) {
			return false;
		}
		utils_printk("Disable PSM: Done\n");
	}

	return true;
}

bool set_fast_release(bool fast_release_enable) {
	if(fast_release_enable == true)
	{
		utils_printk("enable fast release\n");
		if (at_cmd_write("AT%XRAI=4", NULL, 0, NULL) != 0) {
			utils_printk("failed to enable fast release\n");
			return false;
		}
		utils_printk("Success: enable fast release\n");
	}
	else
	{
		utils_printk("disable fast release\n");
		if (at_cmd_write("AT%XRAI=0", NULL, 0, NULL) != 0) {
			utils_printk("failed to disable fast release\n");
			return false;
		}
		utils_printk("Success: disable fast release\n");
	}
	// if (k_sem_take(&registered_on_network_sem, K_SECONDS(REGISTER_ON_NETWORK_TIMEOUT_S)) != 0) {
	// 	utils_printk("set_apn: Did not set fast release in %d seconds.\n", REGISTER_ON_NETWORK_TIMEOUT_S);
	// 	return false;
	// }
	return true;
}

bool send_message(const char *p_message, uint8_t message_length) {
	int sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		utils_printk("Error opening socket: %d\n", errno);
		return false;
	}

	static struct sockaddr_in remote_addr = {
		sin_family: AF_INET,
		sin_port:   htons(1234),
	};
	net_addr_pton(AF_INET, "172.16.15.14", &remote_addr.sin_addr);

	if (sendto(sock, p_message, message_length, 0, (struct sockaddr *)&remote_addr, sizeof(remote_addr)) < 0) {
		utils_printk("Error sending: %d\n", errno);
		close(sock);
		return false;
	}

	close(sock);
	return true;
}

bool m_connect_and_set_psm(void)
{
	utils_printk("Setup to connect the basestation and network.\n");
	if (!subscribe_network_status_notifications()) {
		return false;
	}
    utils_printk("set LET modem mode.\n");
	if (!systemmode_lte()) {
		utils_printk("Failed to set system mode LTE.\n");
		return false;
	}

	// utils_printk("Example application started.\n");

	// if (!print_imei_imsi()) {
	// 	utils_printk("Failed to get IMEI/IMSI.\n");
	// 	return false;
	// }

	// if (!set_apn()) {
	// 	utils_printk("Failed to set APN.\n");
	// 	return false;
	// }
	// utils_printk("Setup to connect the basestation and network: Done.\n");

	// k_sleep(PSM_MODE_SETTING_TIME_AFTER_CONNECTION_S); /* Specifically needed for NBIOT mode */
	// set_psm(PSM_ENBALE);
	// utils_printk("PSM set !\n");

	return true;
}
